"""Confusion matrix functions."""
import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix

from coolpandas.plot import mapplot


def plot_confusion_matrix(cm: np.ndarray) -> None:
    """Plot a confusion matrix.

    Args:
        cm (np.ndarray): Confusion matrix to plot.
    """
    fig = mapplot(cm, title="Confusion matrix", width=400, height=400)
    fig.update_layout(
        xaxis=dict(
            tickmode="array",
            tickvals=[1, 3, 5, 7, 9, 11],
            ticktext=["One", "Three", "Five", "Seven", "Nine", "Eleven"],
        )
    )
    fig.update_layout(
        xaxis_title="Predicted label",
        yaxis_title="True Label",
        xaxis=dict(tickmode="array", tickvals=[0, 1], ticktext=[0, 1]),
        yaxis=dict(tickmode="array", tickvals=[0, 1], ticktext=[0, 1]),
    )
    fig.show()


def get_confusion_matrix(
    y_true: pd.Series | np.ndarray | list,
    y_pred: pd.Series | np.ndarray | list,
    normalize: str | None = None,
    plot: bool = True,
) -> np.ndarray:
    """Create a confusion matrix.
    Args:
        y_true (pd.Series): True labels.
        y_pred (pd.Series): Predicted labels.
        normalize (str, optional): Normalize the confusion matrix.
        Defaults to None.
        plot (bool, optional): Plot the confusion matrix. Defaults to True.

    Returns:
        np.ndarray: Confusion matrix.
    """
    cm: np.ndarray = confusion_matrix(y_true, y_pred, normalize=normalize)
    if plot:
        plot_confusion_matrix(cm)
    return cm



"""Bin a list of values into a list of bins."""
import numpy as np
import pandas as pd


def to_bin(
    series: pd.Series | np.ndarray | list[float | int], step: float = 0.1
) -> pd.Series | np.ndarray | list[float | int]:
    """Bin a series, a numpy array or a list of floats or ints by step.

    Args:
        series (pd.Series | np.ndarray | list[float | int]): Series, numpy array or
        list of floats or ints to bin.
        step (float, optional): Step to bin. Defaults to 0.1.

    Returns:
        pd.Series | np.ndarray | list[float | int]: Binned series, numpy array or list
        of floats or ints.
    """
    binner: callable = lambda x: np.floor(x / step) * step
    return binner(series)



"""Date to epoch transformation."""
import pandas as pd
from pandas._libs.tslibs.timestamps import Timestamp


def timestamp_to_epoch(timestamp: Timestamp | pd.Series) -> int:
    """Convert a timestamp to a unix epoch timestamp.

    Args:
        timestamp (Timestamp | pd.Series): Timestamp or series of timestamps to convert.

    Returns:
        int: Unix epoch timestamp.
    """
    return (timestamp - pd.Timestamp("1970-01-01")) // pd.Timedelta("1s")



"""Style functions for CoolPandas plots."""
import plotly.express as px
import plotly.graph_objects as go

custom_template = {
    "layout": go.Layout(
        font={
            "family": "Nunito",
            "size": 12,
            "color": "#707070",
        },
        title={
            "font": {
                "family": "Lato",
                "size": 18,
                "color": "#1f1f1f",
            },
            "xref": "paper",
            "yref": "paper",
            "xanchor": "left",
            "x": 0,
        },
        plot_bgcolor="#ffffff",
        paper_bgcolor="#ffffff",
        colorway=px.colors.qualitative.G10,
    )
}


def format_title(title: str, subtitle: str = None, subtitle_font_size: int = 14) -> str:
    """Format title and subtitle for plotly figures.

    Args:
        title (str): Title of the plot.
        subtitle (str, optional): Subtitle of the plot. Defaults to None.
        subtitle_font_size (int, optional): Font size of the subtitle. Defaults to 14.

    Returns:
        str: Formatted title.
    """
    title = f"<b>{title}</b>"
    if not subtitle:
        return title
    subtitle = f'<span style="font-size: {subtitle_font_size}px;">{subtitle}</span>'
    return f"{title}<br>{subtitle}"



"""value_counts function for pandas DataFrame."""
import pandas as pd

from coolpandas.plot import barplot


def plot_value_counts(summary: pd.DataFrame) -> None:
    """Plot a DataFrame column value counts overview.

    Args:
        data_frame (pd.DataFrame): DataFrame to get value counts overview.
        column (str): Column to get value counts.
    """
    column: str = summary.columns[0]
    fig = barplot(
        summary,
        x_axis=column,
        y_axis="percentage",
        title=f"{column} value counts",
        hover_data={
            "count": True,
        },
        labels={
            "count": "Count",
            "percentage": "Percentage",
        },
        subtitle=f"Value counts for {summary.columns[0]} in DataFrame.",
        text="count",
    )
    fig.update_layout(
        yaxis={"ticksuffix": "%", "range": [0, 100]},
        xaxis=dict(tickmode="array", tickvals=summary[column].values),
        showlegend=False,
    )
    fig.show()


def get_value_counts(
    data_frame: pd.DataFrame, column: str, plot: bool = True
) -> pd.DataFrame:
    """Get a DataFrame column value counts overview.

    Args:
        data_frame (pd.DataFrame): DataFrame to get value counts overview.
        column (str): Column to get value counts.

    Returns:
        pd.DataFrame: DataFrame value counts summary.
    """
    summary: pd.DataFrame = pd.DataFrame(data_frame[column].value_counts(dropna=False))
    summary.columns = ["count"]
    summary["percentage"] = round(
        data_frame[column].value_counts(dropna=False, normalize=True) * 100, 2
    )
    summary = summary.reset_index().rename(columns={"index": column})
    if plot:
        plot_value_counts(summary)
    return summary


def get_groupby_value_counts(
    data_frame: pd.DataFrame,
    groupby_column: str,
    plot: bool = True,
) -> pd.DataFrame:
    """Get the value counts overview of a DataFrame column grouped by size.

    Args:
        data_frame (pd.DataFrame): DataFrame to get value counts overview.
        groupby_column (str): Column to group by.

    Returns:
        pd.DataFrame: DataFrame value counts summary.
    """
    summary: pd.DataFrame = (
        data_frame.groupby(groupby_column)
        .size()
        .to_frame()
        .reset_index()
        .rename(columns={0: "count"})
    )
    summary["percentage"] = round(summary["count"] / summary["count"].sum() * 100, 2)
    if plot:
        plot_value_counts(summary)
    return summary



"""Outliers transformations."""
import pandas as pd


def iqr_outliers(
    data_frame: pd.DataFrame,
    column: str,
    threshold: float = 1.5,
    create_outlier_column: bool = False,
    lower_bound: float | None = None,
    upper_bound: float | None = None,
) -> pd.DataFrame:
    """Transform outliers using the interquartile range method. If the lower_bound and upper_bound are not provided, they will be calculated using the threshold.

    Args:
        data_frame (pd.DataFrame): DataFrame to get outliers from.
        column (str): Column to get outliers from.
        threshold (float, optional): Threshold to use. Defaults to 1.5.
        create_outlier_column (bool, optional): Whether to create a column with
        outliers. Defaults to False.
        lower_bound (float, optional): Lower bound to use. Defaults to None.
        upper_bound (float, optional): Upper bound to use. Defaults to None.

    Returns:
        pd.DataFrame: DataFrame with outliers.
    """
    if lower_bound is None or upper_bound is None:
        quartile_1: float = data_frame[column].quantile(0.25)
        quartile_3: float = data_frame[column].quantile(0.75)
        iqr: float = quartile_3 - quartile_1
    if lower_bound is None:
        lower_bound: float = quartile_1 - (threshold * iqr)
    if upper_bound is None:
        upper_bound: float = quartile_3 + (threshold * iqr)
    if create_outlier_column:
        data_frame[f"{column}_outlier"] = data_frame[column].apply(
            lambda x: 1 if x < lower_bound or x > upper_bound else 0
        )
    data_frame[column] = data_frame[column].apply(
        lambda x: lower_bound
        if x < lower_bound
        else x
        if x < upper_bound
        else upper_bound
    )
    return data_frame



"""Get a DataFrame column distribution overview."""
import pandas as pd

from coolpandas.plot import distplot


def plot_distribution(
    summary: pd.DataFrame,
    groupby_column: str,
    nbins: int = 100,
    title: str = "",
    subtitle: str = "",
    **kwargs,
) -> None:
    """Plot a feature distribution.

    Args:
        data_frame (pd.DataFrame): DataFrame to get distribution overview.
        groupby_column (str): Column to group by.
        nbins (int, optional): Number of bins. Defaults to 10.
        title (str, optional): Title of the plot. Defaults to "".
        subtitle (str, optional): Subtitle of the plot. Defaults to "".
        **kwargs: Keyword arguments to pass to distplot.
    """
    column: str = summary.columns[0]
    if not title:
        title = f"{column} distribution"
    if not subtitle:
        subtitle = f"Distribution for {summary.columns[0]} in DataFrame."
    if not kwargs.get("color"):
        kwargs["text_auto"] = True
    fig = distplot(
        summary,
        x_axis=groupby_column,
        y_axis=None,
        title=title,
        labels={
            "count": "Count",
        },
        subtitle=subtitle,
        histnorm="percent",
        nbins=nbins,
        **kwargs,
    )
    fig.update_yaxes(tickformat=".3")
    fig.update_layout(
        yaxis={"ticksuffix": "%"},  # , "range": [0, 100]},
        yaxis_title="Percentage",
        barmode="overlay",
    )
    if kwargs.get("color"):
        fig.update_traces(opacity=0.75)
    fig.show()


def distribution_summaries(
    data_frame: pd.DataFrame,
    groupby_column: str,
    data_type: str,
    dependant_value: str = "",
) -> pd.DataFrame:
    if data_type == "categorical":
        summary: pd.DataFrame = (
            data_frame.groupby(groupby_column)
            .size()
            .to_frame()
            .reset_index()
            .rename(columns={0: "count"})
        )
        summary["percentage"] = round(
            summary["count"] / summary["count"].sum() * 100, 2
        )
    elif data_type == "numerical":
        summary = data_frame[groupby_column].describe().to_frame()
    if dependant_value:
        column_name: str = summary.columns[0]
        summary.rename(
            columns={column_name: f"{column_name}_{dependant_value}"}, inplace=True
        )
    return summary


def get_groupby_distribution(
    data_frame: pd.DataFrame,
    groupby_column: str,
    data_type: str,
    plot: bool = True,
    **kwargs,
) -> pd.DataFrame | list[pd.DataFrame]:
    """Get the distribution overview of a DataFrame ordinal column grouped by size.

    Args:
        data_frame (pd.DataFrame): DataFrame to get distribution overview.
        groupby_column (str): Column to group by.
        data_type (str): Data type to get distribution overview.
        plot (bool, optional): Plot the distribution. Defaults to True.
        Defaults to "".
        **kwargs: Keyword arguments to pass to plot_distribution.

    Returns:
        pd.DataFrame | list[pd.DataFrame]: Distribution overviews.
    """
    if plot and data_type == "categorical":
        kwargs["nbins"] = data_frame[groupby_column].nunique()
    elif plot and data_type == "numerical":
        kwargs["marginal"] = "box"
        if not kwargs.get("nbins"):
            kwargs["nbins"] = 31
    if plot and not kwargs.get("title"):
        kwargs["title"] = f"{groupby_column} distribution"
    if plot:
        plot_distribution(
            data_frame,
            groupby_column,
            **kwargs,
        )
    summaries: list[pd.DataFrame] = []
    dependant_column: str = kwargs.get("color")
    if not dependant_column:
        summaries.append(distribution_summaries(data_frame, groupby_column, data_type))
    else:
        summaries = [
            distribution_summaries(
                data_frame[data_frame[dependant_column] == i],
                groupby_column,
                data_type,
                str(i),
            )
            for i in data_frame[dependant_column].unique()
        ]
    return summaries[0] if not dependant_column else summaries



"""Set Python seeds for results reproducibility."""
import os
import random

import numpy as np


def random_state(seed: int = 0) -> None:
    """Set Python seeds for results reproducibility.

    Args:
        seed (int): Seed to use.
    """
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
